#include <string.h>
#include <errno.h>
#include <glib.h>
#include <libguile.h>
#include <vhttpd.h>

/*** main wrappers */
static SCM
scm_from_static_string (const char *str)
{
  return str ? scm_from_utf8_string(str) : SCM_BOOL_F;
}

static SCM
scm_from_dynamic_string(char *s_result)
{
    SCM result = scm_from_static_string(s_result);
    g_free(s_result);
    return result;
}


static void
_assert_number (SCM num, unsigned int arg, const char *func)
{
  SCM_ASSERT (SCM_NUMBERP (num), num, arg, func);
}

static void
_assert_procedure (SCM proc, unsigned int arg, const char *func)
{
  SCM_ASSERT ((SCM_NFALSEP (scm_procedure_p (proc))), proc, arg, func);
}

/*** vhttpd/socket.h */

static SCM
_make_unix_client_socket(SCM path)
{
  char *c_path;
  int s;
  SCM result = SCM_UNDEFINED;

  scm_dynwind_begin(0);

  c_path = scm_to_utf8_stringn(path, NULL);
  scm_dynwind_free (c_path);

  s = create_unix_socket (c_path, 0);
  if (s < 0)
    scm_syserror("make-unix-client-socket");
  else
    result = scm_fdes_to_port(s,"r+0",SCM_BOOL_F);

  scm_dynwind_end();
  return result;
}

static SCM
_make_unix_server_socket(SCM path,SCM user, SCM group)
{
  char *c_path,*c_user,*c_group;
  int s;
  SCM result = SCM_UNDEFINED;

  scm_dynwind_begin(0);

  c_path = scm_to_utf8_stringn(path, NULL);
  scm_dynwind_free (c_path);

  c_user= scm_to_utf8_stringn(user, NULL);
  scm_dynwind_free (c_user);

  c_group= scm_to_utf8_stringn(group, NULL);
  scm_dynwind_free (c_group);

  s = create_unix_socket (c_path, C_SO_FATAL|C_SO_SERVER);
  if (s < 0)
    scm_syserror("make-unix-server-socket");
  else
    result = scm_fdes_to_port(s,"r+0",SCM_BOOL_F);
  unix_socket_set_ownership(c_path, c_user, c_group);

  scm_dynwind_end();
  return result;
}

/*** vhttpd/http_message.h */
static scm_t_bits _message_tag;

static SCM
scm_from_message (const message_t * m)
{
  if (m)
    SCM_RETURN_NEWSMOB (_message_tag, m);
  else
    return SCM_BOOL_F;
}

static message_t *
message_from_scm (SCM m)
{
  return (message_t *) SCM_SMOB_DATA (m);
}

static int
scm_c_message_p (SCM m)
{
  return SCM_SMOB_PREDICATE (_message_tag, m);
}

static void
_assert_message (SCM m, unsigned int arg, const char *func)
{
  SCM_ASSERT (scm_c_message_p (m), m, arg, func);
}

static SCM
_message_p (SCM m)
{
  return scm_c_message_p (m) ? SCM_BOOL_T : SCM_BOOL_F;
}


static void
_message_set_cookies (message_t * r, SCM rest)
{
  SCM cookies = rest;

  if (scm_ilength (cookies) <= 0)
    return;

  while (scm_is_pair (cookies))
    {
      SCM item = SCM_CAR (cookies);
      if (scm_is_pair (item) &&
	  scm_is_string (SCM_CAR (item)) && scm_is_string (SCM_CDR (item)))
	{
	  char *c_i1,*c_i2;
	  scm_dynwind_begin(0);

	  c_i1 = scm_to_utf8_stringn(SCM_CAR(item), NULL);
	  scm_dynwind_free(c_i1);

	  c_i2 = scm_to_utf8_stringn(SCM_CDR(item), NULL);
	  scm_dynwind_free(c_i2);

	  message_set_cookie (r, c_i1, c_i2);

	  scm_dynwind_end();
	}
      cookies = SCM_CDR (cookies);
    }
}


static SCM
_message_get_plain (SCM message)
{
  _assert_message (message, SCM_ARG1, "message-plain?");

  return message_get_plain (message_from_scm (message)) ? SCM_BOOL_T :
    SCM_BOOL_F;
}

static SCM
_message_get_header (SCM message, SCM header)
{
  _assert_message (message, SCM_ARG1, "message-header");

  char *c_header;
  SCM result;
  scm_dynwind_begin(0);

  c_header = scm_to_utf8_stringn(header, NULL);
  scm_dynwind_free(c_header);
  const char *r = message_get_header (message_from_scm (message), c_header);

  result = scm_from_static_string (r);

  scm_dynwind_end();
  return result;
}

static SCM
_message_set_header (SCM message, SCM header,SCM value)
{
  _assert_message (message, SCM_ARG1, "message-header");

  char *c_header,*c_value;
  scm_dynwind_begin(0);

  c_header = scm_to_utf8_stringn(header, NULL);
  scm_dynwind_free(c_header);
  c_value = scm_to_utf8_stringn(value, NULL);
  scm_dynwind_free(c_value);

  message_set_header (message_from_scm (message), c_header, c_value);

  scm_dynwind_end();
  return SCM_UNSPECIFIED;
}

static SCM
_message_get_cookie (SCM message, SCM name)
{
  _assert_message (message, SCM_ARG1, "message-cookie");

  char *c_name;
  SCM result;

  scm_dynwind_begin(0);

  c_name = scm_to_utf8_stringn(name, NULL);
  scm_dynwind_free(c_name);

  const char *r = message_get_cookie (message_from_scm (message), c_name);
  result = scm_from_static_string(r);

  scm_dynwind_end();
  return result;
}


static SCM
_message_get_method (SCM message)
{
  _assert_message (message, SCM_ARG1, "message-method");
  return scm_from_static_string (message_get_method (message_from_scm (message)));
}

static SCM
_message_get_uri (SCM message)
{
  _assert_message (message, SCM_ARG1, "message-uri");
  return scm_from_static_string (message_get_uri (message_from_scm (message)));
}

static SCM
_message_get_query (SCM message)
{
  _assert_message (message, SCM_ARG1, "message-uri");
  return scm_from_static_string (message_get_query (message_from_scm (message)));
}

static SCM
_message_get_body_main (void *data)
{
  message_t *m = (message_t *) data;

  const char *body = message_get_body(m);
  int len = message_get_length(m);

  if (len <= 0)
    return scm_from_utf8_stringn("", 0);
  else
    return scm_from_utf8_stringn(body, len);
}

static SCM
_message_get_body_fallback (void *data, SCM key, SCM args)
{
  message_t *m = (message_t *) data;

  const char *body = message_get_body(m);
  int len = message_get_length(m);

  if (len <= 0)
    return scm_from_latin1_stringn("", 0);
  else
    return scm_from_latin1_stringn(body, len);
}

static SCM
_message_get_body (SCM message)
{
  _assert_message (message, SCM_ARG1, "message-body");
  message_t *m = message_from_scm (message);
  return scm_internal_catch (SCM_BOOL_T,
                             _message_get_body_main, m,
                             _message_get_body_fallback, m);
}

static SCM
_message_get_startline (SCM message)
{
  _assert_message (message, SCM_ARG1, "message-startline");
  return scm_from_static_string (message_get_startline (message_from_scm (message)));
}

static SCM
_message_get_code (SCM message)
{
  _assert_message (message, SCM_ARG1, "message-code");
  return scm_from_int (message_get_code (message_from_scm (message)));
}

static SCM
_message_get_code_string (SCM code)
{
  _assert_number (code, SCM_ARG1, "message-code-string");
  return scm_from_static_string (http_strerror(scm_to_int(code)));
}

static SCM
_make_response(SCM code)
{
    _assert_number (code, SCM_ARG1, "make-response");
    return scm_from_message(create_response_message(scm_to_int(code)));
}

static SCM
_make_string_response (SCM retcode, SCM type, SCM body, SCM rest)
{
  char *c_type,*c_body;
  message_t *r = NULL;
  SCM result;

  scm_dynwind_begin(0);

  c_type = scm_to_utf8_stringn(type, NULL);
  scm_dynwind_free (c_type);

  c_body = scm_to_utf8_stringn(body, NULL);
  if (!c_body)
    runtime_error("make_string_response: scm_to_utf8_stringn() error");

  r = create_response_message (scm_to_int (retcode));
  message_set_body (r, c_body, strlen(c_body));
  message_set_header (r, "content-type", c_type);
  _message_set_cookies (r, rest);

  result = scm_from_message (r);

  free(c_body);

  scm_dynwind_end();
  return result;
}

static SCM
_make_file_response (SCM path,SCM mtime)
{
  char *c_path,*c_mtime;
  SCM result;
  scm_dynwind_begin(0);

  c_path = scm_to_utf8_stringn(path, NULL);
  scm_dynwind_free (c_path);
  c_mtime = scm_to_utf8_stringn(mtime, NULL);
  scm_dynwind_free (c_mtime);

  message_t *r = message_file_new (c_path,c_mtime);
  result = scm_from_message (r);

  scm_dynwind_end();
  return result;
}

static SCM
_make_redirect_response (SCM location, SCM rest)
{
  char *c_location;
  SCM result;
  scm_dynwind_begin(0);

  c_location = scm_to_utf8_stringn(location, NULL);
  scm_dynwind_free (c_location);

  message_t *r = message_redirect_new (c_location);
  _message_set_cookies (r, rest);
  result = scm_from_message (r);

  scm_dynwind_end();
  return result;
}


/*** vhttpd/daemon.h */

static SCM
_daemonize(SCM pidfile)
{
    char *c_pidfile;

    scm_dynwind_begin(0);

    c_pidfile = scm_to_utf8_stringn(pidfile, NULL);
    scm_dynwind_free (c_pidfile);

    daemonize(c_pidfile);

    scm_dynwind_end();

    return SCM_UNSPECIFIED;
}

static SCM
_drop_privs(SCM user,SCM group)
{
    char *c_user,*c_group;

    scm_dynwind_begin(0);

    c_user = scm_to_utf8_stringn(user, NULL);
    scm_dynwind_free (c_user);

    c_group = scm_to_utf8_stringn(group, NULL);
    scm_dynwind_free (c_group);

    drop_privs(c_user,c_group);

    scm_dynwind_end();

    return SCM_UNSPECIFIED;
}

/*** vhttpd/decode.h */

static SCM
_url_decode(SCM str)
{
    char *c_str;
    SCM result;

    scm_dynwind_begin(0);

    c_str = scm_to_utf8_stringn(str, NULL);
    scm_dynwind_free (c_str);

    result =  scm_from_dynamic_string(url_decode(c_str,strlen(c_str)));

    scm_dynwind_end();

    return result;
}

static SCM
_url_encode(SCM str)
{
    char *c_str;
    SCM result;

    scm_dynwind_begin(0);

    c_str = scm_to_utf8_stringn(str, NULL);
    scm_dynwind_free (c_str);

    result =  scm_from_dynamic_string(url_encode(c_str,strlen(c_str)));

    scm_dynwind_end();

    return result;
}

static SCM
_base64_decode(SCM str)
{
    char *c_str;
    SCM result;

    scm_dynwind_begin(0);

    c_str = scm_to_utf8_stringn(str, NULL);
    scm_dynwind_free (c_str);

    result = scm_from_dynamic_string(base64_decode(c_str,0));

    scm_dynwind_end();

    return result;
}

/*** vhttpd/http_server.h */

static scm_t_bits _server_tag;

static SCM
scm_from_server (const server_t * c)
{
  if (c)
    SCM_RETURN_NEWSMOB (_server_tag, c);
  else
    return SCM_BOOL_F;
}

static server_t *
server_from_scm (SCM c)
{
  return (server_t *) SCM_SMOB_DATA (c);
}

static int
scm_c_server_p (SCM c)
{
  return SCM_SMOB_PREDICATE (_server_tag, c);
}

static void
_assert_server (SCM c, unsigned int arg, const char *func)
{
  SCM_ASSERT (scm_c_server_p (c), c, arg, func);
}

static SCM
_server_p (SCM c)
{
  return scm_c_server_p (c) ? SCM_BOOL_T : SCM_BOOL_F;
}

static SCM
_make_server_from_socket (SCM socket_fd)
{
    SCM result;

    result = scm_from_server(create_server_from_socket(scm_to_int(socket_fd)));

    return result;
}

static SCM
_make_tls_server_from_socket (SCM socket_fd, SCM keyfile, SCM certfile)
{
    char *c_keyfile,*c_certfile;
	int c_fd;
    SCM result;

    scm_dynwind_begin(0);

    c_fd = scm_to_int(socket_fd);

    c_keyfile = scm_to_utf8_stringn(keyfile, NULL);
    scm_dynwind_free(c_keyfile);

    c_certfile = scm_to_utf8_stringn(certfile, NULL);
    scm_dynwind_free(c_certfile);

    result = scm_from_server(create_smart_server_from_socket(c_fd, c_keyfile, c_certfile));

    scm_dynwind_end();
    return result;
}

static SCM
_make_tls_server (SCM addr, SCM port, SCM keyfile, SCM certfile)
{
    char *c_addr,*c_port,*c_keyfile,*c_certfile;
    SCM result;

    scm_dynwind_begin(0);

    c_addr = scm_to_utf8_stringn(addr, NULL);
    scm_dynwind_free(c_addr);

    c_port = scm_to_utf8_stringn(port, NULL);
    scm_dynwind_free(c_port);

    c_keyfile = scm_to_utf8_stringn(keyfile, NULL);
    scm_dynwind_free(c_keyfile);

    c_certfile = scm_to_utf8_stringn(certfile, NULL);
    scm_dynwind_free(c_certfile);

    result = scm_from_server(create_smart_server(c_addr, c_port, c_keyfile, c_certfile));

    scm_dynwind_end();
    return result;
}

static SCM
_make_unix_server (SCM path)
{
    char *c_path;
    SCM result;

    scm_dynwind_begin(0);

    c_path = scm_to_utf8_stringn(path, NULL);
    scm_dynwind_free(c_path);

    result = scm_from_server(create_unix_server(c_path));

    scm_dynwind_end();
    return result;
}

static message_t *
_message_handler (retcode_t retcode, const message_t * request,void *data)
{
  SCM handler = (SCM)data;
  SCM response = scm_apply_2 (handler,
			      scm_from_int ((int) retcode),
			      scm_from_message (request),
			      SCM_EOL);

  return scm_c_message_p (response) ? message_from_scm (response) : NULL;
}

static SCM
_server_loop(SCM server, SCM handler)
{
    _assert_server(server,SCM_ARG1,"message-loop");
    _assert_procedure(handler,SCM_ARG2,"message-loop");

    server_loop (server_from_scm(server), _message_handler, handler);

    return SCM_UNSPECIFIED;
}

/*** vhttpd/http_server.h */

static SCM
_request_unix_server(SCM path, SCM body, SCM handler)
{
    _assert_procedure(handler,SCM_ARG3,"message-loop");

    message_t *request = create_request_message("GET", "/");
    message_t *response = NULL;
    retcode_t retcode;
    char *c_path,*c_body;
    SCM result;

    scm_dynwind_begin(0);

    c_path = scm_to_utf8_stringn(path, NULL);
    scm_dynwind_free (c_path);

    c_body = scm_to_utf8_stringn(body, NULL);
    if (!c_body)
	runtime_error("request_unix_server: scm_to_utf8_stringn() error");

    message_set_body(request, c_body, strlen(c_body));
    retcode = request_unix_server(c_path, request, &response);

    result = scm_apply_2(handler,
                         scm_from_int ((int) retcode),
                         scm_from_message (response),
                         SCM_EOL);

    destroy_message(request);
    destroy_message(response);
    free(c_body);

    scm_dynwind_end();
    return result;
}

/*** functions */

void scm_init_vhttpd (void);

void
scm_init_vhttpd (void)
{
  /* vhttpd/socket.h */
  scm_c_define_gsubr ("make-unix-client-socket", 1, 0, 0, _make_unix_client_socket);
  scm_c_define_gsubr ("make-unix-server-socket", 3, 0, 0, _make_unix_server_socket);

  /* vhttpd/http_message.h */
  _message_tag = scm_make_smob_type ((char *) "message", 0);
  scm_c_define_gsubr ("message?", 1, 0, 0, _message_p);
  scm_c_define_gsubr ("message-plain?", 1, 0, 0, _message_get_plain);
  scm_c_define_gsubr ("message-header", 2, 0, 0, _message_get_header);
  scm_c_define_gsubr ("message-header-set!", 3, 0, 0, _message_set_header);
  scm_c_define_gsubr ("message-cookie", 2, 0, 0, _message_get_cookie);
  scm_c_define_gsubr ("message-method", 1, 0, 0, _message_get_method);
  scm_c_define_gsubr ("message-uri", 1, 0, 0, _message_get_uri);
  scm_c_define_gsubr ("message-query", 1, 0, 0, _message_get_query);
  scm_c_define_gsubr ("message-body", 1, 0, 0, _message_get_body);
  scm_c_define_gsubr ("message-startline", 1, 0, 0, _message_get_startline);
  scm_c_define_gsubr ("message-code", 1, 0, 0, _message_get_code);
  scm_c_define_gsubr ("message-code-string", 1, 0, 0, _message_get_code_string);
  scm_c_define_gsubr ("make-response", 1, 0, 1, _make_response);

  scm_c_define_gsubr ("make-string-response", 3, 0, 1, _make_string_response);
  scm_c_define_gsubr ("make-file-response", 2, 0, 0, _make_file_response);
  scm_c_define_gsubr ("make-redirect-response", 1, 0, 1, _make_redirect_response);

  /* vhttpd/daemon.h */
  scm_c_define_gsubr ("daemonize", 1, 0, 0, _daemonize);
  scm_c_define_gsubr ("drop-privs", 2, 0, 0, _drop_privs);

  /* vhttpd/decode.h */
  scm_c_define_gsubr ("decode-url-component", 1, 0, 0, _url_decode);
  scm_c_define_gsubr ("encode-url-component", 1, 0, 0, _url_encode);
  scm_c_define_gsubr ("decode-base64", 1, 0, 0, _base64_decode);

  /* vhttpd/http_server.h */
  _server_tag = scm_make_smob_type ((char *) "server", 0);
  scm_c_define_gsubr ("server?", 1, 0, 0, _server_p);
  scm_c_define_gsubr ("make-server-from-socket", 1, 0, 0, _make_server_from_socket);
  scm_c_define_gsubr ("make-tls-server-from-socket", 3, 0, 0, _make_tls_server_from_socket);
  scm_c_define_gsubr ("make-tls-server", 4, 0, 0, _make_tls_server);
  scm_c_define_gsubr ("make-unix-server", 1, 0, 0, _make_unix_server);
  scm_c_define_gsubr ("server-loop", 2, 0, 0, _server_loop);

  /* vhttpd/http_client.h */
  scm_c_define_gsubr ("request-unix-server", 3, 0, 0, _request_unix_server);
}
